import os
from pathlib import Path

RBX_ENV = os.getenv("RBX_ENV", "dev")
RBX_PROJECT = f"{RBX_ENV}-platform-eu"

GOOGLE_CLOUD_PROJECT = os.getenv("GOOGLE_CLOUD_PROJECT")

MDM_PUBSUB_TOPIC = os.getenv("MDM_PUBSUB_TOPIC", "platform-notifications")
MDM_PAYLOAD_VERSION = 2

AWS_BUCKET = os.getenv("AWS_BUCKET")
AWS_REGION = os.getenv("AWS_REGION")

PROJECT_ROOT = Path(__file__).resolve().parent.parent



from copy import copy
from functools import partial
import json
import logging
from urllib.parse import urljoin

import requests
from requests.auth import AuthBase
from requests.exceptions import ConnectionError, Timeout

from ..exceptions import (
    ClientError,
    Forbidden,
    ServerError,
    TransientServerError,
    Unauthorized,
)
from ..utils import Singleton

logger = logging.getLogger(__name__)


class HttpAuth(AuthBase):
    """Attaches HTTP AUTH-TOKEN Authentication to the given Request object."""

    def __init__(self, token, key="X-AUTH-TOKEN"):
        self.key = key
        self.token = token

    def __call__(self, r):
        r.headers[self.key] = self.token
        return r

    def __eq__(self, other):
        return all(
            [
                other.key == self.key,
                other.token == self.token,
            ]
        )


class Client(metaclass=Singleton):
    """Base Client offering basic functionality, such as logging in, auto-login on expiry, and
    common error handling.

    The Client is a JSON client, meaning it is to be used with an API that supports JSON format
    in its requests and responses.

    This class must be extended, do not use it directly.
    """

    AUTH_PATH = "auth/login"
    DEFAULT_TIMEOUT = 30
    ENDPOINT = "/"
    TOKEN = "token"

    def __init__(self):
        self.credentials = None
        self.token = None

    @property
    def auth(self):
        """Authentication method.

        Returns None by default, meaning no authentication.
        One can include Digest Authentication using the HttpAuth object, e.g.:

        >>> return HttpAuth(self.token, key='X-AUTH')

        """
        return None

    @property
    def is_authenticated(self):
        return self.token is not None

    def get_message(self, response):
        """Extract the message from the requests response object."""
        return self.get_response(response).get("message")

    def get_response(self, response):
        """Extract the response payload from the requests response object."""
        return response.json()

    def login(self, username, password):
        """Authenticate and set the access token on success."""
        response = self._request(
            "post",
            self.AUTH_PATH,
            data={
                "username": username,
                "password": password,
            },
        )

        self.credentials = {
            "username": username,
            "password": password,
        }
        self.token = response.get(self.TOKEN)
        logger.debug(f"Logged in <Token: {self.token}>")

    def logout(self):
        self.token = None
        if hasattr(self, "csrftoken"):
            self.csrftoken = None

    def refresh_access_token(self):
        raise NotImplementedError

    def _post(self, path, **kwargs):
        """Shortcut for _request('post')."""
        return self._request("post", path, **kwargs)

    def _request(
        self,
        method,
        path,
        content_type="application/json",
        data=None,
        endpoint=None,
        headers=None,
        params=None,
        timeout=None,
    ):
        """Wrap the method call in common error handling."""
        endpoint = endpoint if endpoint is not None else self.ENDPOINT
        url = urljoin(endpoint, path)

        # Default requests parameter (except login).
        args = {"timeout": self.DEFAULT_TIMEOUT if timeout is None else timeout}
        if self.auth:
            args["auth"] = self.auth

        # CSRF Cookies hack required by some vendors (i.e.: Broadsign).
        if hasattr(self, "csrftoken") and self.csrftoken and method in ("post", "put"):
            args.update(
                {
                    "headers": {"X-CSRFToken": self.csrftoken},
                    "cookies": {"csrftoken": self.csrftoken},
                }
            )

        # Add custom headers.
        if headers:
            if "headers" in args:
                args["headers"].update(headers)
            else:
                args["headers"] = headers
        # Add params
        if params is not None:
            args["params"] = params

        # Use the right requests parameter according to the method and data type.
        if data is not None:
            if method == "get":
                args["params"] = args.get("params", {}) | data
            elif type(data) is str:
                args["data"] = data
            else:
                # If the data payload can't be serialized into JSON, then pass it as provided.
                # Otherwise, pass is a as JSON value.
                try:
                    json.dumps(data)
                    args["json"] = data
                except TypeError:
                    args["data"] = data

        try:
            if path == self.AUTH_PATH:
                # Don't use authentication for the token endpoint.
                response = requests.post(url, json=data, timeout=self.DEFAULT_TIMEOUT)
            else:
                response = getattr(requests, method)(url, **args)
        except (ConnectionError, Timeout) as e:
            raise TransientServerError(message=str(e), status_code=500, url=url)

        # Never include the password field in the log on failure
        payload = copy(data)
        if payload and type(payload) is dict:
            payload.pop("password", None)

        if response.status_code == 500:
            raise TransientServerError(response.text, details=payload, url=url)

        if response.status_code == 401:
            raise Unauthorized(self.get_message(response), details=payload, url=url)

        if response.status_code == 403:
            raise Forbidden(self.get_message(response), details=payload, url=url)

        if response.status_code >= 400:
            raise ClientError(
                response.text,
                details=payload,
                status_code=response.status_code,
                url=url,
            )

        # CSRF Cookies hack, continued. Store the first 'csrftoken' value found from the response
        # cookies. The Set-Cookie won't be sent on login, so we have to extract it from the first
        # GET request.
        if (
            hasattr(self, "csrftoken")
            and "csrftoken" in response.cookies
            and method == "get"
        ):
            self.csrftoken = response.cookies["csrftoken"]

        # 204 "No Content" responses return nothing
        if response.status_code == 204:
            return

        if content_type == "text/plain":
            return response.text
        else:
            try:
                return self.get_response(response)
            except ValueError:
                extra = {
                    "request": payload,
                    "response": response.text,
                }
                raise ServerError(
                    "Invalid JSON response", details=extra, status_code=500, url=url
                )

    def request(
        self,
        method,
        path,
        content_type="application/json",
        data=None,
        endpoint=None,
        headers=None,
        params=None,
        timeout=None,
    ):
        """Attempt to re-authenticate when the token has expired.

        Note that the Client must be logged in first for this to work.
        """
        if not self.is_authenticated:
            raise Unauthorized

        api_call = partial(
            self._request,
            method=method,
            path=path,
            content_type=content_type,
            data=data,
            endpoint=endpoint,
            headers=headers,
            params=params,
            timeout=timeout,
        )

        try:
            return api_call()

        except Unauthorized as e:
            # The token has expired. If the client implements the `refresh_access_token` method,
            # then attempt to get a fresh token that way.
            # Otherwise, start fresh.
            logger.debug(f'Access denied: "{e.status_code} - {e.message}"')
            try:
                self.refresh_access_token()
            except (Forbidden, Unauthorized, NotImplementedError):
                self.logout()
                self.login(**self.credentials)

            return api_call()



from unittest import mock

from .keystore import Key, Keystore


def fake_key(
    key_id="6DSfBxFKsp3EDfUAETZn",
    campaigns=None,
    email="dev-mode@localhost.com",
    is_restricted=True,
    name=None,
    services=None,
    status="active",
):
    return Key(
        keystore=mock.Mock(spec_set=Keystore),
        key_id=key_id,
        campaigns=campaigns or [],
        email=email or "john.doe@rip.com",
        is_restricted=is_restricted,
        name=name or "John Doe",
        services=services if services is not None else ["*.*"],
        status=status,
    )


class KeystoreMock:
    def add_campaign(self, *args, **kwargs):
        return

    def add_campaigns(self, *args, **kwargs):
        return

    def create_key(
        self, email, name, campaigns=None, is_restricted=None, services=None
    ):
        return fake_key(
            key_id="6DSfBxFKsp3EDfUAETZn",
            name=name,
            email=email or "dev-mode@localhost.com",
            is_restricted=False,
        )

    def get_campaigns(self):
        return []

    def get_key(self, key_id=None, email=None, name=None):
        return fake_key(
            key_id=key_id or "6DSfBxFKsp3EDfUAETZn",
            name=name or "DEV Mode",
            email=email or "dev-mode@localhost.com",
            is_restricted=False,
        )

    def update_key(self, *args, **kwargs):
        return



from functools import wraps
from typing import NamedTuple

from flask import abort, current_app, request

from . import Keystore
from .mock import KeystoreMock


class Grants(NamedTuple):
    key: str
    campaigns: list = []
    is_restricted: bool = False


def protect(with_grants=False):
    """Protect an endpoint via API Keys.

    If with_grants is True, the Grants object is added to the caller as a 'grants' keyword
    argument.

    When DEV_MODE is enabled, all restrictions are lifted.

    Keys are retrieved from the Keystore, and cached using the LRU local caching.
    They are provided via the "X-RBX-TOKEN" header, and can also be provided via the 'key'
    query string parameter.

    If the request originates from App Engine's Cron Service, or from Cloud Tasks, it is always
    granted access.

    From the documentation:

        The X-Appengine-Cron header is set internally by Google App Engine.
        If your request handler finds this header it can trust that the request is a cron request.
        The X- headers are stripped by App Engine when they originate from external sources so that
        you can trust this header.

    """

    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            if any(
                [
                    "X-Appengine-Cron" in request.headers,
                    "X-AppEngine-TaskName" in request.headers,
                ]
            ):
                return func(*args, **kwargs)

            key_id = request.headers.get("X-RBX-TOKEN") or request.args.get("key")
            if key_id is None:
                abort(401)

            # Use the Mock implementation when DEV_MODE is enabled or if we're using the
            # Google "mock" project.
            if (
                current_app.config.get("DEV_MODE", False)
                or current_app.config.get("GOOGLE_CLOUD_PROJECT") == "mock"
            ):
                impl = KeystoreMock
            else:
                impl = Keystore

            key = impl().get_key(key_id=key_id)

            qualname = f"{func.__module__}.{func.__name__}"
            endpoint = next(
                (
                    rule
                    for rule in current_app.url_map.iter_rules()
                    if qualname.endswith(rule.endpoint)
                ),
                None,
            )
            if not endpoint:
                abort(401)

            service = current_app.name.replace("_", "-")
            operation = str(endpoint).lstrip("/").replace("/", ".")

            if not key or not key.has_access(service=service, operation=operation):
                abort(401)

            if with_grants:
                kwargs["grants"] = Grants(
                    key=key_id, campaigns=key.campaigns, is_restricted=key.is_restricted
                )

            return func(*args, **kwargs)

        return wrapped

    return decorator



from dataclasses import asdict, dataclass, field, InitVar
import os
from unittest import mock

import google.auth.credentials
from google.api_core import retry
from google.api_core.exceptions import RetryError
from google.cloud.firestore import Client

from ..clients.retry import Retry
from ..exceptions import TransientException
from ..settings import RBX_PROJECT

default_retry = retry.Retry(initial=0.1, maximum=3.0, multiplier=1.3, deadline=10.0)


class Keystore:
    """Manages API Keys stored in Google Cloud Firestore (Native Mode)."""

    def __init__(self):
        # Use mocked credentials to interact with the Firestore emulator.
        if os.getenv("GAE_ENV", "").startswith("standard"):
            db = Client(project=RBX_PROJECT)
        else:
            credentials = mock.Mock(spec=google.auth.credentials.Credentials)
            db = Client(
                project=os.getenv("GOOGLE_CLOUD_PROJECT"), credentials=credentials
            )

        self.collection = db.collection("api_keys")
        self.inventory = db.collection("inventory")

    def _inventory(self, key, document):
        """Given a DocumentSnapshot, return the fully loaded Inventory object."""
        return Inventory(**{"key": key, "keystore": self, **document.to_dict()})

    def _key(self, document):
        """Given a DocumentSnapshot, return the fully loaded Key object."""
        return Key(**{"keystore": self, "key_id": document.id, **document.to_dict()})

    def add_campaign(self, campaign):
        """Add a campaign to the campaigns inventory."""
        self.add_campaigns([campaign])

    @Retry(deadline=120.0)
    def add_campaigns(self, campaigns):
        """Add campaigns to the campaigns inventory."""
        inventory = self.get_inventory()
        inventory.add(campaigns)

    @Retry(deadline=120.0)
    def create_key(
        self, email, name, campaigns=None, is_restricted=None, services=None
    ):
        """Make a new API key.

        A key is associated with a user via her email address. If a key already exists for that
        user, the key is updated instead.
        """
        key = self.get_key(email=email)
        if not key:
            key = {
                "campaigns": campaigns or [],
                "email": email,
                "is_restricted": is_restricted if is_restricted is not None else True,
                "name": name,
                "services": services if services is not None else ["*.*"],
                "status": "active",
            }

            try:
                _, ref = self.collection.add(key, retry=default_retry)
                key = self._key(ref.get(retry=default_retry))
            except RetryError as e:
                raise TransientException(e)
        else:
            key.update(
                **{
                    "campaigns": campaigns or key.campaigns,
                    "is_restricted": (
                        is_restricted
                        if is_restricted is not None
                        else key.is_restricted
                    ),
                    "name": name or key.name,
                    "services": services or key.services,
                    "status": "active",  # the key is resuscitated
                }
            )

        # Unrestricted keys get access to all inventory
        if not key.is_restricted:
            key.campaigns = self.get_campaigns()

        return key

    def get_campaigns(self):
        """List all campaigns in the inventory."""
        inventory = self.get_inventory()
        if not inventory:
            return []

        return inventory.values

    def get_inventory(self, key="campaigns"):
        """Retrieve a key from the inventory."""
        try:
            document = self.inventory.document(key).get(retry=default_retry)
        except RetryError as e:
            raise TransientException(e)

        if not document.exists:
            return Inventory(key=key, keystore=self)
        else:
            return self._inventory(key, document)

    def get_key(self, key_id=None, email=None):
        """Retrieve a key by ID or email."""
        key = None

        try:
            if key_id:
                document = self.collection.document(key_id).get(retry=default_retry)
                if document.exists:
                    key = self._key(document)

            if email:
                document = next(
                    self.collection.where("email", "==", email).stream(
                        retry=default_retry
                    ),
                    False,
                )
                if document:
                    key = self._key(document)
        except RetryError as e:
            raise TransientException(e)

        # Unrestricted keys get access to all inventory
        if key and not key.is_restricted:
            key.campaigns = self.get_campaigns()

        return key

    def update_key(self, key, attributes):
        """Set the new Key values in Firestore."""
        assert isinstance(key, Key), f"excepted rbx.auth.Key, got {type(key)}"
        document = self.collection.document(key.key_id)
        try:
            document.set(attributes, merge=True, retry=default_retry)
        except RetryError as e:
            raise TransientException(e)


@dataclass
class Inventory:
    key: InitVar[str]
    keystore: InitVar[Keystore]
    values: list = field(default_factory=list)

    def __post_init__(self, key, keystore):
        self.key = key
        self.keystore = keystore

    def add(self, added_values):
        values = set(self.values)
        values.update(set(added_values))
        document = self.keystore.inventory.document(self.key)
        try:
            document.set(
                {"values": sorted(list(values))}, merge=True, retry=default_retry
            )
        except RetryError as e:
            raise TransientException(e)


@dataclass
class Key:
    key_id: InitVar[str]
    keystore: InitVar[Keystore]
    email: str
    name: str
    campaigns: list = field(default_factory=list)
    is_restricted: bool = True
    services: list = field(default_factory=["*.*"])
    status: str = "active"

    def __post_init__(self, key_id, keystore):
        # These are defined as InitVar so that they are not part of the pickled data, and aren't
        # included in the to_dict() representation.
        self.key_id = key_id
        self.keystore = keystore

    def activate(self):
        self.update(status="active")

    def deactivate(self):
        self.update(status="inactive")

    def has_access(self, service, operation):
        """Check whether the service and operation are granted access by this Key."""
        if self.status != "active":
            return False

        operation_ns = operation.split(".")
        for grant in self.services:
            name, _, endpoint = grant.partition(".")

            if name == "*":
                return True

            elif name == service:
                try:
                    if all(
                        [
                            value == "*" or value == operation_ns[key]
                            for key, value in enumerate(endpoint.split("."))
                        ]
                    ):
                        return True

                except IndexError:
                    continue

        return False

    def to_dict(self):
        return asdict(self)

    def update(self, **kwargs):
        for key, value in kwargs.items():
            if key in self.__annotations__.keys() and key not in ("key_id", "keystore"):
                setattr(self, key, value)

        self.keystore.update_key(self, self.to_dict())



import click

from .tasks import apprunner, ec2, image, misc


from rbx import __version__


@click.group(
    context_settings={
        "help_option_names": ["-h", "--help"],
    },
    add_help_option=True,
)
@click.version_option(message=__version__)
def cli():
    """Scoota (a.k.a. RBX) buildtools."""


cli.add_command(ec2.deploy)


@cli.group(name="apprunner")
def apprunner_group():
    """Manage AppRunner services."""


apprunner_group.add_command(apprunner.submit)
apprunner_group.add_command(apprunner.deploy)


@cli.group(name="image")
def image_group():
    """Manage images on Amazon Container Registry."""


image_group.add_command(image.login)
image_group.add_command(image.build)
image_group.add_command(image.upload)


@cli.group(name="misc")
def misc_group():
    """Miscellaneous tools."""


misc_group.add_command(misc.clean)
misc_group.add_command(misc.get_next_tag)
misc_group.add_command(misc.get_merge_desc)
misc_group.add_command(misc.get_npm_version)



import logging

from . import Client, HttpAuth

logger = logging.getLogger(__name__)


class PanelsClient(Client):
    """API Client for the Panels Service."""

    def __init__(self, endpoint, token):
        super().__init__()
        self.ENDPOINT = endpoint.rstrip("/")
        self.token = token

    @property
    def auth(self):
        """The Panels Service uses Digest Authentication."""
        return HttpAuth(self.token, key="X-RBX-TOKEN")

    def search(self, **parameters):
        """Perform a Panels search."""
        return self._post("/search", data=parameters)

    def set_weather_watch(self, caller, country, supplier):
        """Set a watch on the specified country and supplier, using the caller ID."""
        return self.request(
            method="get",
            path="/forecast/watch",
            data={"caller": caller, "country": country, "supplier": supplier},
        )

    def set_weather_unwatch(self, caller, country, supplier):
        """Remove a watch on the specified country and supplier, using the caller ID."""
        return self.request(
            method="get",
            path="/forecast/unwatch",
            data={"caller": caller, "country": country, "supplier": supplier},
        )



"""Common utility methods."""

__all__ = ["Singleton", "smart_url"]

import re
from threading import Lock
from urllib.parse import urljoin


class Singleton(type):
    """Metaclass providing a thread-safe Singleton interface."""

    _instances = {}
    _singleton_lock = Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._singleton_lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


def smart_url(url, force_https=False, prefix=None):
    """Return the appropriate URL.

    The returned URL will be:
     - an empty string, when the provided url is None;
     - the provided url, unchanged, when it includes the protocol ('http', 'https' or '//');
     - otherwise, the provided url, optionally prefixed with the given prefix.
     - finally, if force_https is set, we ensure the URL is return with the HTTPS protocol.
    """
    if not url:
        return ""

    if not force_https and (url.startswith("http") or url.startswith("//")):
        return url

    if prefix:
        url = urljoin(prefix, url)

    if force_https:
        match = re.match("http:|https:", url)
        if match:
            _, _, url = url.partition(match.group(0))

        url = urljoin("https://", url)

    return url



import subprocess
from os import path

import click


@click.command()
def clean(ctx):
    """Remove all intermediate files from the source tree."""
    subprocess.run("git clean -xdf", shell=True)


@click.command(name="git-next-tag")
def get_next_tag():
    """Generate tag from Git.

    A helper to generate the next tag based on the current git tag.

    To use in Jenkins job:

        >>> export VERSION=`buildtools misc git-next-tag`

    Expects an integer, otherwise will start at 1.
    """
    cmd = "git describe --tags $(git rev-list --tags --max-count=1)"
    process = subprocess.run(cmd, shell=True, capture_output=True)
    try:
        tag = int(process.stdout.strip())
    except ValueError:
        tag = 0
    print(tag + 1)


@click.command(name="git-merge-desc")
def get_merge_desc():
    """Get description from Git.

    A helper to get the description of a merge commit.

    To use in a Jenkins job:

        >>> export DESCRIPTION=`buildtools misc git-merge-desc`

    When merging a PR the subject line will be added to the body of the merge
    commit.
    """
    cmd = "git log -1 --pretty='format:%s'"
    process = subprocess.run(cmd, shell=True, capture_output=True)
    print(process.stdout.decode().strip())


@click.command(name="npm-version")
def get_npm_version():
    """Get version from NPM.

    A helper to get the npm package version from a package.json file.

    To use in Jenkins job:

        >>> export VERSION=`buildtools misc npm-version`

    """
    if path.exists("package.json"):
        cmd = "cat package.json | jq .version -r"
        process = subprocess.run(cmd, shell=True, capture_output=True)
        print(process.stdout.decode().strip())
    else:
        print("unknown")



import subprocess
from pathlib import Path
from textwrap import dedent

import click
from jinja2 import Environment

from .. import REGION, REGISTRY


@click.command()
@click.argument("app")
@click.option("-s", "--source", default=".")
@click.option("-v", "--version", default="latest")
def submit(app, source, version):
    """Submit a local Docker build and push the image to AWS ECR."""
    tag = f"{REGISTRY}/{app}:{version}"

    # Login to the Amazon Container Registry
    subprocess.run(
        "aws ecr get-login-password | docker login"
        f" --username AWS --password-stdin {REGISTRY}",
        shell=True,
    )

    # Build image
    subprocess.run(f"docker build -t {tag} {source}", shell=True)

    # Push image
    subprocess.run(f"docker push {tag}", shell=True)


@click.command()
@click.argument("app")
@click.option("--env", multiple=True)
@click.option("--secret", multiple=True)
@click.option("--name")
@click.option("--port", type=int, default=8000)
@click.option("-s", "--source", default=".")
@click.option("-v", "--version", default="latest")
def deploy(ctx, app, env, secret, name, port, source, version):
    """Deploy an image to App Runner."""
    config_file = Path("apprunner.yaml")
    source_file = Path(source) / "apprunner.yaml"

    if not source_file.exists():
        raise click.Abort("Cannot find source apprunner.yaml file")

    if source_file != config_file and config_file.exists():
        raise click.Abort(
            "Conflicting apprunner.yaml files (found file both at root and source)"
        )

    delete_on_exit = source_file != config_file

    with source_file.open() as fp:
        config = fp.read()

    with config_file.open("w") as fp:
        fp.write(config)

    if delete_on_exit:
        config_file.unlink()

    environments = dict([tuple(kv.split("=")) for kv in env])
    secrets = dict([tuple(kv.split("=")) for kv in secret])

    template = Environment(lstrip_blocks=True).from_string(
        dedent(
            """\
            {
              "ImageRepository": {
                "ImageIdentifier": "{{tag}}",
                "ImageRepositoryType": "ECR",
                "ImageConfiguration": {
                  "Port": "{{port}}",
                  "RuntimeEnvironmentVariables": {
                    {%- for key, value in environments.items() %}
                    "{{key}}": "{{value}}"{% if loop.last %}{% else %},{% endif %}
                    {%- endfor %}
                  },
                  "RuntimeEnvironmentSecrets": {
                    {%- for key, value in secrets.items() %}
                    "{{key}}": "{{value}}"{% if loop.last %}{% else %},{% endif %}
                    {%- endfor %}
                  }
                }
              },
              "AutoDeploymentsEnabled": false,
              "AuthenticationConfiguration": {
                "AccessRoleArn": "arn:aws:iam::474071279654:role/Service"
              }
            }
            """
        )
    )

    configuration = template.render(
        environments=environments,
        port=port,
        secrets=secrets,
        tag=f"{REGISTRY}/{app}:{version}",
    )

    name = name or app

    # Does the service already exist?
    query = f"ServiceSummaryList[?ServiceName == '{name}'].ServiceArn | [0]"
    command = subprocess.run(
        f'aws apprunner list-services --query "{query}"',
        shell=True,
        capture_output=True,
    )
    arn = command.stdout.strip().strip('"')

    if arn == "null":
        # Create service
        subprocess.run(
            "aws apprunner create-service"
            f" --region {REGION}"
            f" --service-name {name}"
            f" --source-configuration '{configuration}'"
            " --instance-configuration 'InstanceRoleArn=arn:aws:iam::474071279654:role/Service'"
            " --no-cli-pager",
            shell=True,
        )
    else:
        # Update service
        subprocess.run(
            "aws apprunner update-service"
            f" --region {REGION}"
            f" --service-arn {arn}"
            f" --source-configuration '{configuration}'"
            " --no-cli-pager",
            shell=True,
        )



import subprocess
from os import getcwd, path

import click

from .. import REGISTRY

BASEPATH = path.basename(getcwd())


@click.command()
def login():
    """Login to the Amazon Container Registry."""
    subprocess.run(
        "aws ecr get-login-password | docker login"
        f" --username AWS --password-stdin {REGISTRY}",
        shell=True,
    )


@click.command()
@click.option("--build-arg", multiple=True)
@click.option("--cache/--no-cache", default=True)
@click.option("--name", default=BASEPATH)
@click.option("--version", default="develop")
@click.option("--target")
@click.option("--test/--no-test", default=False)
def build(build_arg, cache, name, version, target, test):
    """Build the docker image.

    Note that parameters always take precedence over environment variables.
    """
    command = ["docker build"]

    if target:
        command.append("--target {}".format(target))

    if test:
        command.append("--build-arg INSTALL_TEST_DEPENDENCIES=yes")

    for arg in build_arg:
        command.append("--build-arg {}".format(arg))

    if not cache:
        command.append("--no-cache")

    command.append(f"-t {REGISTRY}/{name}:{version} .")
    command = " ".join(command)
    subprocess.run(command, shell=True)


@click.command()
@click.option("--name", default=BASEPATH)
@click.option("--version", default="develop")
def upload(name, version):
    """Upload the docker image to Amazon Container Registry.

    Needless to say, this image must have been already built.

    Note that parameters always take precedence over environment variables.
    """
    subprocess.run(f"docker push {REGISTRY}/{name}:{version}", shell=True)



import logging
from urllib.parse import quote, parse_qs, urlparse

import arrow
from cachetools import cached, TTLCache

from ..utils import smart_url
from . import Client, HttpAuth

logger = logging.getLogger(__name__)


class BroadsignClient(Client):
    """Client for Broadsign Reach APIs.

    The API documentation can be found on Google Drive:
    https://drive.google.com/file/d/17APVeWVlVLf4T8tqXiexWzfaVnh0U9Oo/view

    """

    AUTH_ENDPOINT = "https://auth.broadsign.com"
    AUTH_PATH = "/oauth/token"
    CLIENT_ID = "dIJ179dH1A2Wn9j9ia4rYGICNmuJMMYb"
    ENDPOINT = "https://servssp.broadsign.com/"
    IF_MODIFIED_SINCE_FORMAT = "ddd, DD MMM YYYY HH:mm:ss [GMT]"
    TOKEN = "access_token"
    REFRESH_TOKEN = "refresh_token"

    def __init__(self, dsp_id, dsp_name):
        super().__init__()
        self.csrftoken = None
        self.dsp_id = dsp_id
        self.dsp_name = dsp_name
        self.refresh_token = None

    @property
    def auth(self):
        """Broadsign uses Bearer Authentication."""
        return HttpAuth(f"Bearer {self.token}", key="Authorization")

    def login(self, username, password):
        """Authenticate and set the access token and refresh token on success.

        If the Client is already authenticated, the login process will be skipped.
        """
        if self.is_authenticated:
            return

        response = self._request(
            "post",
            self.AUTH_PATH,
            endpoint=self.AUTH_ENDPOINT,
            data={
                "username": username,
                "password": password,
                "grant_type": "http://auth0.com/oauth/grant-type/password-realm",
                "client_id": self.CLIENT_ID,
                "realm": "Username-Password-Authentication",
                "audience": "https://platform.broadsign.com/",
                "scope": "offline_access",
            },
        )

        self.credentials = {
            "username": username,
            "password": password,
        }
        self.token = response.get(self.TOKEN)
        self.refresh_token = response.get(self.REFRESH_TOKEN)
        logger.debug(
            f"Logged in <Token: {self.token}> <Refresh Token: {self.refresh_token}>"
        )

    def logout(self):
        super().logout()
        self.csrftoken = None
        self.refresh_token = None

    def get_message(self, response):
        return response.text

    def refresh_access_token(self):
        """Retrieve access token and refresh token from Broadsign with previous refresh token."""
        response = self.request(
            "post",
            self.AUTH_PATH,
            endpoint=self.AUTH_ENDPOINT,
            data={
                "grant_type": "refresh_token",
                "client_id": self.CLIENT_ID,
                "refresh_token": self.refresh_token,
            },
        )

        self.token = response.get(self.TOKEN)
        self.refresh_token = response.get(self.REFRESH_TOKEN)
        logger.debug(
            f"Refreshed <Token: {self.token}> <Refresh Token: {self.refresh_token}>"
        )

    def create_advertiser(self, name):
        """Create an Advertiser object."""
        response = self.request(
            "post",
            "/api/advertisers/",
            data={
                "name": name,
                "dsp": {"id": self.dsp_id, "name": self.dsp_name},
            },
        )
        return {
            "id": response["id"],
            "name": response["name"],
        }

    def create_creative(
        self,
        advertiser,
        creative_id,
        category,
        creative_type,
        creative_url,
        name,
        publishers,
    ):
        """Create a Creative object.

        Note that the 'original_url' requires the protocol, or the API request will fail.
        """
        creative_types = {
            "static": "ImageUrlCreative",
            "video": "VideoCreative",
            "html": "Html5Creative",
        }

        response = self.request(
            "post",
            "/api/entity_creatives/",
            data={
                "advertiser": advertiser,
                "external_id": str(creative_id),
                "iab_categories": [category],
                "name": name,
                "original_url": smart_url(creative_url, force_https=True),
                "publishers": publishers,
                "type": creative_types.get(creative_type, "ImageUrlCreative"),
            },
        )

        return {
            "id": int(response["external_id"]),
            "name": response["name"],
            "status": response["status"],
            "type": response["type"],
        }

    def get_advertiser(self, name):
        """Retrieve an Advertiser object by name."""
        path = "/api/advertisers/?name={}".format(quote(name))
        response = self.request("get", path)
        if response.get("count", 0) > 0:
            return {
                "id": response["results"][0]["id"],
                "name": response["results"][0]["name"],
            }

    def get_category(self, code):
        """Retrieve a Category object by code."""
        path = "/api/categories/?code={}".format(code)
        response = self.request("get", path)
        if response.get("count", 0) > 0:
            return {
                "id": response["results"][0]["id"],
                "name": response["results"][0]["name"],
            }

    def get_creative(self, creative_id, as_entity=False):
        """Retrieve a Creative object by ID.

        Passing the as_entity flag as True will return the object as returned by the system.
        """
        path = "/api/entity_creatives/?external_id={}".format(creative_id)
        response = self.request("get", path)
        if response.get("count", 0) > 0:
            entity = response["results"][0]
            if as_entity:
                return entity

            return {
                "id": int(entity["external_id"]),
                "name": entity["name"],
                "status": entity["status"],
                "type": entity["type"],
            }

    @cached(cache=TTLCache(maxsize=4, ttl=3600))
    def get_publishers(self, ad_type=None):
        """Return the full list of publishers.

        This method is cached for 1 day.
        """
        return self._get_publishers(ad_type=ad_type)

    def _get_publishers(self, ad_type=None, page=None):
        """Retrieve a list of Publisher objects at the given page.

        As the API restrict the response to 100 items, we need to repeat the process until we have
        reached the last page.

        The list may be further refined by ad_type, one of 'html', 'static', or 'video'.
        The API does not allow filtering by query string, so we have to resort to filtering the
        final results ourselves.

        Note that when a publisher has no ad types defined (i.e.: 'allowed_ad_types': []), we
        include the publisher. According to Broadsign, this happens when the details on the
        publisher haven't made it to the API yet, so we should add them.
        """
        if ad_type:
            assert ad_type in (
                "html",
                "static",
                "video",
            ), f'Invalid Ad Type: "{ad_type}"'

        path = "/api/publishers/"
        params = {"page_size": 100}
        if page:
            params["page"] = page

        publishers = []
        response = self.request("get", path, data=params)
        if response.get("count", 0) > 0:
            publishers.extend(response["results"])

        if response.get("next"):
            try:
                page = int(parse_qs(urlparse(response.get("next")).query)["page"][0])
            except (KeyError, ValueError):
                pass
            else:
                publishers.extend(self._get_publishers(page=page))

        if ad_type:
            publishers[:] = [
                publisher
                for publisher in publishers
                if not publisher["allowed_ad_types"]
                or ad_type
                in [
                    allowed_ad_type["name"].lower()
                    for allowed_ad_type in publisher["allowed_ad_types"]
                ]
            ]

        publishers.sort(key=lambda x: x["name"])

        return publishers

    def get_screen(self, last_modified=None, page=1):
        """Retrieve a page of screens from Broadsign

        Parameters:
            last modified: date
                Restrict the results to panels entered past the specified date.
            page: int
                The page number of the panels to retrieve.

        Returns:
            tuple: a tuple containing a list of screens retrieved from broadsign, and the next page.
            If there are no further pages, page will be set to -1.
        """
        path = "/api/screens/"
        params = {"page_size": 100}
        if page:
            params["page"] = page

        headers = None

        if last_modified:
            utc = arrow.get(last_modified).to("utc")
            headers = {"If-Modified-Since": utc.format(self.IF_MODIFIED_SINCE_FORMAT)}

        screens = []
        response = self.request("get", path, data=params, headers=headers)
        if response.get("count", 0) > 0:
            screens.extend(response["results"])

        # assume there are no further pages by default
        page = None
        if response.get("next"):
            try:
                page = int(parse_qs(urlparse(response.get("next")).query)["page"][0])
            except (KeyError, ValueError):
                pass

        return screens, page

    def get_screens(self, last_modified=None):
        """Retrieve all screens from Broadsign.

        The list may be further refined using the last_modified date, which will cause the API to
        return only those screens that were modified after the specified date.
        """

        screens = []
        page = 1
        while True:
            result, page = self.get_screen(last_modified=last_modified, page=page)
            if result:
                screens.extend(result)
            if not page:
                break

        return screens

    def update_creative(self, entity_id, payload):
        """Update a Creative object.

        Note that updating requires the entity_id, which is the ID of the Creative in the supplier
        system.
        """
        self.request("put", f"/api/entity_creatives/{entity_id}/", data=payload)



import click
from fabric import Connection

from .. import CONFIGURATIONS, REGISTRY


@click.command()
@click.argument("host")
@click.argument("user")
@click.argument("key_filename")
@click.argument("service")
@click.argument("environment")
@click.option("--sandboxed/--no-sandboxed", default=True)
@click.option("-v", "--version", default="latest")
def deploy(host, user, key_filename, service, environment, sandboxed, version):
    """Deploy to EC2.

    Download a Docker image and start the container on a Remote Server.
    """
    connection = Connection(
        host=host, user=user, connect_kwargs={"key_filename": key_filename}
    )

    # Login to Scoota Docker Registry
    connection.run(
        "aws ecr get-login-password | docker login"
        f" --username AWS --password-stdin {REGISTRY}",
        echo=True,
    )

    # Get the latest configuration for the Server we are deploying to.
    # Valid configurations are located at '/<Service>/<Env>/'
    connection.run(
        "aws s3 sync {config}/{service}/{env} {service}".format(
            config=CONFIGURATIONS, env=environment, service=service
        ),
        echo=True,
    )

    # Restart the Docker container forcing recreate so that new version is used.
    # Using `force-recreate` alone doesn't seem to be enough to use the latest version. So we
    # do `down` and then `up`.
    connection.run(
        "docker-compose -p {service}{env} -f {service}/docker-compose.yml down"
        " --volumes --remove-orphans".format(service=service, env=environment),
        echo=True,
        env={
            "SANDBOXED": sandboxed,
            "VERSION": version,
        },
        warn=True,
    )
    connection.run(
        "docker-compose -p {service}{env} -f {service}/docker-compose.yml up -d".format(
            service=service, env=environment
        ),
        echo=True,
        env={
            "SANDBOXED": sandboxed,
            "VERSION": version,
        },
    )

    # Remove all unused Docker Images
    connection.run("docker rmi `docker images -q` || true", echo=True)
    connection.run("docker rm `docker ps -a -q` || true", echo=True)



import logging
from typing import NamedTuple

import arrow
import requests
from requests.exceptions import ConnectionError, Timeout

logger = logging.getLogger(__name__)


class ExchangeRate(NamedTuple):
    """Represents an exchange rate published at a specific time."""

    name: str
    published_at: arrow.Arrow
    value: float

    def __repr__(self) -> str:
        return f"<ExchangeRate: ({self})>"

    def __str__(self) -> str:
        return f"{self.name}: {self.value}"


class OXRClient:
    """A simple HTTP client to retrieve currency exchange rates.

    See https://docs.openexchangerates.org/docs/latest-json for the implementation doc.
    """

    LATEST_ENDPOINT = "https://openexchangerates.org/api/latest.json"
    TIMEOUT = 5

    def __init__(self, app_id):
        self.app_id = app_id

    def latest(self, currencies):
        """Get the latest exchange rates available from the Open Exchange Rates API.

        Note that all rates are based on USD and this base currency is also given in the rates
        object by default.

        Usage:
        >>> client = OXRClient(app_id='92db86.........')
        >>> client.latest(['GBP', 'EUR'])
        [<ExchangeRate (EUR: 0.876002)>, <ExchangeRate (GBP: 0.788209)>, <ExchangeRate (USD: 1)>]

        """
        result = self._get(url=self.LATEST_ENDPOINT, currencies=currencies)
        try:
            published_at = arrow.get(result.get("timestamp"))
        except TypeError:
            return []

        return [
            ExchangeRate(name=name, published_at=published_at, value=value)
            for name, value in result.get("rates", {}).items()
        ]

    def _get(self, url, currencies):
        """Get currency exchange rates.

        The default base currency of the Open Exchange Rates API is US Dollars (USD).
        """
        if not currencies:
            logger.error("At least 1 currency is required.")
            return {}

        params = {
            "app_id": self.app_id,
            "symbols": ",".join(currencies),
            "timeout": self.TIMEOUT,
        }

        try:
            response = requests.get(url, params=params)
        except (ConnectionError, Timeout):
            logger.error(
                "Error connecting to OXR",
                extra={
                    "params": params,
                    "url": url,
                },
            )
            return {}

        if response.status_code >= 400 and response.status_code < 500:
            logger.error(
                response.json()["description"],
                extra={
                    "params": params,
                    "response": response.json(),
                    "url": url,
                },
            )
            return {}

        if response.status_code != 200:
            logger.error(
                response.text,
                extra={
                    "params": params,
                    "response": response.json(),
                    "url": url,
                },
            )
            return {}

        return response.json()



__all__ = [
    "AdParameters",
    "build_vpaid",
    "build_vast",
    "Creative",
    "MediaFile",
    "replace_cachebusting",
    "Tracker",
]

import dataclasses as dc
import json
import re
from typing import Literal, Optional, TypedDict, Union

from lxml import etree


@dc.dataclass
class MediaFile:
    url: str
    mimetype: str
    width: int
    height: int
    bitrate: Optional[int] = 0
    delivery: Literal["progressive"] = "progressive"


@dc.dataclass
class Creative:
    id: str
    name: str
    description: str
    duration: int
    files: [MediaFile]
    icon: Optional[bool] = False
    skipoffset: Optional[int] = 0


@dc.dataclass
class Wrapper:
    id: str
    uri: str


class Attribute(TypedDict):
    name: str
    value: str


class EventParameter(TypedDict):
    pixel: list[str]
    redirect: list[str]
    script: list[str]


class EventsParameters(TypedDict):
    clicks: list[str]
    engagement: EventParameter
    impression: EventParameter


class TrackingParameters(TypedDict):
    events: EventsParameters
    params: dict


class AdParameters(TypedDict):
    dcoMacros: dict
    tracking: TrackingParameters


@dc.dataclass
class Tracker:
    type: Literal["click", "clickthrough", "impression", "error", "event"]
    url: str
    attributes: Optional[list[Attribute]] = None

    def __hash__(self):
        return hash(
            (self.type, self.url, tuple(self.attributes) if self.attributes else None)
        )


def build_vast(
    *,
    id: str,
    creative: Union[Creative, Wrapper],
    trackers: Optional[list[Tracker]] = None,
    version: Optional[str] = "3.0",
) -> bytes:
    """Construct a VAST XML.

    Note that all trackers support multiple occurences except "clickthrough".
    """
    is_wrapper = isinstance(creative, Wrapper)
    has_video = (
        False
        if is_wrapper
        else any(file.mimetype.startswith("video") for file in creative.files)
    )

    tracking = {
        "errors": [],
        "impressions": [],
        "events": [],
        "clickthrough": None,
        "click": None,
    }
    for tracker in trackers or []:
        if tracker.type == "error":
            tracking["errors"].append(tracker)
        elif tracker.type == "impression":
            tracking["impressions"].append(tracker)
        elif tracker.type == "event":
            # VAST 2.0 does not support "closeLinear", "progress" and "skip" events.
            if version == "2.0" and tracker.attributes["event"] in (
                "closeLinear",
                "progress",
                "skip",
            ):
                continue
            # VAST Audio does not need "fullscreen" and "exitFullscreen" events.
            if not has_video and tracker.attributes["event"] in (
                "fullscreen",
                "exitFullscreen",
            ):
                continue
            tracking["events"].append(tracker)
        elif tracker.type == "clickthrough":
            tracking["clickthrough"] = tracker
        elif tracker.type == "click":
            tracking["click"] = tracker

    """Root element

    <VAST xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
        xsi:noNamespaceSchemaLocation="vast.xsd"
        version="3.0">
        <Ad id="Placement ID">
            ...
        </Ad>
    </VAST>

    """
    ns_xsi = "http://www.w3.org/2001/XMLSchema-instance"
    root = etree.Element(
        "VAST",
        attrib={
            f"{{{ns_xsi}}}noNamespaceSchemaLocation": "vast.xsd",
            "version": version,
        },
    )

    ad_elmt = etree.SubElement(root, "Ad")
    ad_elmt.set("id", id)

    """<Inline> element

        <InLine>
            <AdSystem>System</AdSystem>
            <AdTitle>Title</AdTitle>
            <Description><![CDATA[Description]]></Description>
            <Error><![CDATA[...]]></Error>
            <Impression><![CDATA[...]]></Impression>
        </InLine>

    <AdSystem>, <AdTitle> and <Impression> are required elements.

    or <Wrapper> element

        <Wrapper>
            <AdSystem>System</AdSystem>
            <VASTAdTagURI><![CDATA[...]]></VASTAdTagURI>
            <Error><![CDATA[...]]></Error>
            <Impression><![CDATA[...]]></Impression>
        </Wrapper>

    <AdSystem> and <Impression> are required elements.
    """
    if is_wrapper:
        inline_elmt = etree.SubElement(ad_elmt, "Wrapper")
    else:
        inline_elmt = etree.SubElement(ad_elmt, "InLine")

    ad_system_elmt = etree.SubElement(inline_elmt, "AdSystem")
    ad_system_elmt.text = "Scoota"

    if is_wrapper:
        ad_tag_uri_elmt = etree.SubElement(inline_elmt, "VASTAdTagURI")
        ad_tag_uri_elmt.text = etree.CDATA(creative.uri)

    else:
        ad_title_elmt = etree.SubElement(inline_elmt, "AdTitle")
        ad_title_elmt.text = creative.name

        description_elmt = etree.SubElement(inline_elmt, "Description")
        description_elmt.text = creative.description

    # <Error/>
    for tracker in tracking["errors"]:
        error_elmt = etree.SubElement(inline_elmt, "Error")
        error_elmt.text = etree.CDATA(tracker.url)

    # <Impression/>
    for tracker in tracking["impressions"]:
        imp_elmt = etree.SubElement(
            inline_elmt, "Impression", attrib=tracker.attributes
        )
        imp_elmt.text = etree.CDATA(tracker.url)

    """<Creatives> element.

    The <Creatives> element provides details about the files for each placement to be included
    as part of the ad experience.

        <Creatives>
            <Creative id="Creative ID">
                <Linear>
                    <Duration>HH:MM:SS.mmm</Duration>
                    <MediaFiles>
                        <MediaFile
                            delivery="progressive"
                            type="video/mp4"
                            width="300"
                            height="250">
                            <![CDATA[http://path/to/video.mp4]]>
                        </MediaFile>
                    </MediaFiles>
                    <TrackingEvents>
                        <Tracking><![CDATA[...]]></Tracking>
                    </TrackingEvents>
                    <VideoClicks>
                        <ClickThrough><![CDATA[...]]></ClickThrough>
                        <ClickTracking><![CDATA[...]]></ClickTracking>
                    </VideoClicks>
                    <Icon />
                </Linear>
            </Creative>
        </Creatives>

    A <Linear> element has two required child elements, <Duration> and <MediaFiles>.

    `delivery`, `type`, `width` and `height` are all required attributes of <MediaFile>. The
    `delivery` attribute defaults to progressive as we currently do not support streaming video.
    """
    has_tracking = len(tracking["events"]) > 0 or tracking["clickthrough"] is not None

    if not is_wrapper or has_tracking:
        # A wrapper with tracking needs the <Linear> element to exist
        creatives_elmt = etree.SubElement(inline_elmt, "Creatives")
        creative_elmt = etree.SubElement(
            creatives_elmt, "Creative", attrib={"id": creative.id}
        )
        linear_elmt = etree.SubElement(creative_elmt, "Linear")

    if not is_wrapper:
        if creative.skipoffset:
            linear_elmt.set(
                "skipoffset", milliseconds_to_HHMMSSmmm(creative.skipoffset)
            )

        # <Duration> element.
        duration_elmt = etree.SubElement(linear_elmt, "Duration")
        duration_elmt.text = milliseconds_to_HHMMSSmmm(creative.duration)

        # Required <MediaFiles> element.
        media_files_elmt = etree.SubElement(linear_elmt, "MediaFiles")
        for file in creative.files:
            attrib = {
                "delivery": file.delivery,
                "height": str(file.height),
                "maintainAspectRatio": "true",
                "scalable": "true",
                "type": file.mimetype,
                "width": str(file.width),
            }
            if file.bitrate:
                attrib["bitrate"] = str(file.bitrate)

            media_file_elmt = etree.SubElement(
                media_files_elmt,
                "MediaFile",
                attrib=attrib,
            )
            media_file_elmt.text = etree.CDATA(file.url)

    """Tracking <TrackingEvents> and <VideoClicks> elements.

        <TrackingEvents>
            <Tracking event="firstQuartile">
                <![CDATA[http://path/to/first/quartile]]>
            </Tracking>
        </TrackingEvents>
        <VideoClicks>
            <ClickThrough>
                <![CDATA[http://path/to/clickthrough]]>
            </ClickThrough>
            <ClickTracking>
                <![CDATA[http://path/to/click]]>
            </ClickTracking>
        </VideoClicks>

    """
    if len(tracking["events"]):
        events_elmt = etree.SubElement(linear_elmt, "TrackingEvents")
        for tracker in tracking["events"]:
            event_elmt = etree.SubElement(
                events_elmt, "Tracking", attrib=tracker.attributes
            )
            event_elmt.text = etree.CDATA(tracker.url)

    if tracking["clickthrough"]:
        tracker = tracking["clickthrough"]
        video_clicks_elmt = etree.SubElement(linear_elmt, "VideoClicks")
        clickthrough_elmt = etree.SubElement(
            video_clicks_elmt, "ClickThrough", attrib=tracker.attributes
        )
        clickthrough_elmt.text = etree.CDATA(tracker.url)
    else:
        video_clicks_elmt = None

    if tracking["click"]:
        tracker = tracking["click"]
        if video_clicks_elmt is None:
            video_clicks_elmt = etree.SubElement(linear_elmt, "VideoClicks")
        click_elmt = etree.SubElement(
            video_clicks_elmt, "ClickTracking", attrib=tracker.attributes
        )
        click_elmt.text = etree.CDATA(tracker.url)

    """AdChoices <Icon> element.

    Display AdChoices icon and will click through to //info.evidon.com/more_info/130210.
    """
    if not is_wrapper and creative.icon:
        icons_elmt = etree.SubElement(linear_elmt, "Icons")
        icon_elmt = etree.SubElement(
            icons_elmt,
            "Icon",
            attrib={
                "height": "15",
                "program": "AdChoices",
                "width": "77",
                "xPosition": "right",
                "yPosition": "top",
            },
        )
        static_resource_elmt = etree.SubElement(
            icon_elmt,
            "StaticResource",
            attrib={
                "creativeType": "image/png",
            },
        )
        static_resource_elmt.text = etree.CDATA("//c.betrad.com/icon/c_30_us.png")
        icon_clicks_elmt = etree.SubElement(icon_elmt, "IconClicks")
        icon_clickthrough_elmt = etree.SubElement(icon_clicks_elmt, "IconClickThrough")
        icon_clickthrough_elmt.text = etree.CDATA("//info.evidon.com/more_info/130210")

    return etree.tostring(root, encoding="UTF-8", xml_declaration=False)


def build_vpaid(
    *,
    id: str,
    creative: Creative,
    ad_parameters: Optional[AdParameters] = None,
    trackers: Optional[list[Tracker]] = None,
    version: Optional[str] = "3.0",
) -> bytes:
    """Construct a VPAID VAST XML."""
    tracking = {"errors": [], "events": [], "clickthrough": None}
    for tracker in trackers or []:
        if tracker.type == "error":
            tracking["errors"].append(tracker)
        elif tracker.type == "event":
            # VAST 2.0 does not support "progress" and "skip" events.
            if version == "2.0" and tracker.attributes["event"] in ("progress", "skip"):
                continue
            tracking["events"].append(tracker)
        elif tracker.type == "clickthrough":
            tracking["clickthrough"] = tracker

    """Root element

    <VAST xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
        xsi:noNamespaceSchemaLocation="vast.xsd"
        version="3.0">
        <Ad id="Placement ID">
            ...
        </Ad>
    </VAST>

    """
    ns_xsi = "http://www.w3.org/2001/XMLSchema-instance"
    root = etree.Element(
        "VAST",
        attrib={
            f"{{{ns_xsi}}}noNamespaceSchemaLocation": "vast.xsd",
            "version": version,
        },
    )

    ad_elmt = etree.SubElement(root, "Ad")
    ad_elmt.set("id", id)

    """<Inline> element

        <InLine>
            <AdSystem>System</AdSystem>
            <AdTitle>Title</AdTitle>
            <Description><![CDATA[Description]]></Description>
            <Impression><![CDATA[...]]></Impression>
        </InLine>

    <AdSystem>, <AdTitle> and <Impression> are required elements.
    """
    inline_elmt = etree.SubElement(ad_elmt, "InLine")
    ad_system_elmt = etree.SubElement(inline_elmt, "AdSystem")
    ad_system_elmt.text = "Scoota"

    ad_title_elmt = etree.SubElement(inline_elmt, "AdTitle")
    ad_title_elmt.text = creative.name

    description_elmt = etree.SubElement(inline_elmt, "Description")
    description_elmt.text = creative.description

    # Dummy <Impression/>
    imp_elmt = etree.SubElement(inline_elmt, "Impression")
    imp_elmt.text = etree.CDATA("")

    # <Error/>
    for tracker in tracking["errors"]:
        error_elmt = etree.SubElement(inline_elmt, "Error")
        error_elmt.text = etree.CDATA(tracker.url)

    """<Creatives> element.

        <Creatives>
            <Creative id="Creative ID">
                <Linear>
                    <Duration>HH:MM:SS.mmm</Duration>
                    <MediaFiles>
                        <MediaFile
                            apiFramework="VPAID"
                            delivery="progressive"
                            type="application/javascript"
                            width="300"
                            height="250">
                            <![CDATA[http://path/to/vpaid.js]]>
                        </MediaFile>
                    </MediaFiles>
                    <AdParameters><![CDATA[...]]></AdParameters>
                </Linear>
            </Creative>
        </Creatives>

    """
    creatives_elmt = etree.SubElement(inline_elmt, "Creatives")
    creative_elmt = etree.SubElement(
        creatives_elmt, "Creative", attrib={"id": creative.id}
    )
    linear_elmt = etree.SubElement(creative_elmt, "Linear")

    if creative.skipoffset:
        linear_elmt.set("skipoffset", milliseconds_to_HHMMSSmmm(creative.skipoffset))

    # <Duration> element.
    duration_elmt = etree.SubElement(linear_elmt, "Duration")
    duration_elmt.text = milliseconds_to_HHMMSSmmm(creative.duration)

    # VPAID <MediaFiles> element.
    file = creative.files[0]
    media_files_elmt = etree.SubElement(linear_elmt, "MediaFiles")
    media_file_elmt = etree.SubElement(
        media_files_elmt,
        "MediaFile",
        attrib={
            "apiFramework": "VPAID",
            "delivery": file.delivery,
            "height": str(file.height),
            "type": file.mimetype,
            "width": str(file.width),
        },
    )
    media_file_elmt.text = etree.CDATA(file.url)

    """Tracking <TrackingEvents> and <VideoClicks> elements.

        <TrackingEvents>
            <Tracking event="firstQuartile">
                <![CDATA[http://path/to/first/quartile]]>
            </Tracking>
        </TrackingEvents>
        <VideoClicks>
            <ClickThrough>
                <![CDATA[http://path/to/clickthrough]]>
            </ClickThrough>
        </VideoClicks>

    """
    if len(tracking["events"]):
        events_elmt = etree.SubElement(linear_elmt, "TrackingEvents")
        for tracker in tracking["events"]:
            event_elmt = etree.SubElement(
                events_elmt, "Tracking", attrib=tracker.attributes
            )
            event_elmt.text = etree.CDATA(tracker.url)

    if tracking["clickthrough"]:
        tracker = tracking["clickthrough"]
        video_clicks_elmt = etree.SubElement(linear_elmt, "VideoClicks")
        clickthrough_elmt = etree.SubElement(
            video_clicks_elmt, "ClickThrough", attrib=tracker.attributes
        )
        clickthrough_elmt.text = etree.CDATA(tracker.url)

    # <AdParameters> element.
    if ad_parameters:
        ad_parameters_elmt = etree.SubElement(linear_elmt, "AdParameters")
        ad_parameters_elmt.text = etree.CDATA(json.dumps(ad_parameters))

    return etree.tostring(root, encoding="UTF-8", xml_declaration=False)


def milliseconds_to_HHMMSSmmm(milliseconds: int) -> str:
    """Convert milliseconds to HH:MM:SS.mmm string format."""
    hours, rem = divmod(milliseconds / 1000.0, 3600)
    minutes, seconds = divmod(rem, 60)

    return "{:0>2}:{:0>2}:{:06.3f}".format(int(hours), int(minutes), seconds)


def replace_cachebusting(url):
    """Replace [timestamp] and [cachebuster] macros with [CACHEBUSTING] macro."""
    return re.sub(
        pattern=r"\[(timestamp|cachebuster)\]",
        repl="[CACHEBUSTING]",
        string=url,
        flags=re.I,
    )
